/*
 * _mssql module
 *
 * Initial Developer:
 *      jooncheol@gmail.com <Joon-cheol Park >, http://www.exman.pe.kr
 *
 * Active Developer:
 *      Andrzej Kukula <akukula@gmail.com>
 *
 * thankyou:
 *     Hans Roh (hans@lufex.com) - List ref count patch 2004.04.09
 * */

#include <Python.h>
#include <structmember.h>

#ifdef MS_WINDOWS
  #define DBNTWIN32	// must identify operating system environment
  #define NOCRYPT	// must be defined under Visual C++ .NET 2003
  #include <windows.h>
  #include "odbcinst.h"	// SQLConfigDataSource and ODBC_ADD_DSN
  #include <lmerr.h>
  #include <sqlfront.h>
  #include <sqldb.h>	// DB-LIB header file (should always be included)
#else
  #define MSDBLIB	// we need FreeTDS to provide MSSQL API,
  			// not Sybase API. See README.freetds for details
  #include <sqlfront.h>
  #include <sqldb.h>

  #define SQLNUMERIC	SYBNUMERIC
  #define SQLDECIMAL	SYBDECIMAL
  #define SQLUNIQUE	SYBUNIQUE
  #define SQLBIT	SYBBIT
  #define SQLINT1	SYBINT1
  #define SQLINT2	SYBINT2
  #define SQLINT4	SYBINT4
  #define SQLINTN	SYBINTN
  #define SQLFLT4	SYBREAL
  #define SQLFLT8	SYBFLT8
  #define SQLFLTN	SYBFLTN
  #define SQLDATETIME	SYBDATETIME
  #define SQLDATETIM4	SYBDATETIME4
  #define SQLDATETIMN	SYBDATETIMN
  #define SQLMONEY	SYBMONEY
  #define SQLMONEY4	SYBMONEY4
  #define SQLMONEYN	SYBMONEYN
  #define SQLBINARY	SYBBINARY
  #define SQLVARBINARY	SYBVARBINARY
  #define SQLIMAGE	SYBIMAGE
  #define SQLVARCHAR	SYBVARCHAR
  #define SQLCHAR	SYBCHAR
  #define SQLTEXT	SYBTEXT

  #define BYTE		unsigned char
  typedef unsigned char *LPBYTE;
#endif

#define TYPE_STRING	1
#define TYPE_BINARY	2
#define TYPE_NUMBER	3
#define TYPE_DATETIME	4
#define TYPE_ROWID	5

#include <stdio.h>
#include <string.h>	// include for string functions
#include <stdlib.h>	// include for malloc and free

static PyObject *_mssql_error;

// Prototypes for internal functions.
PyObject *GetRow(DBPROCESS *dbproc);
PyObject *GetHeaders(DBPROCESS *dbproc);

// Connection object
typedef struct {
  PyObject_HEAD
  DBPROCESS *dbproc;	//PDBPROCESS dbproc;
  int connected;
} _mssql_ConnectionObj;

static PyTypeObject _mssql_ConnectionObj_Type;

#define MSSQLDB_MSGSIZE 1024
static char _mssql_query_error[MSSQLDB_MSGSIZE]={0,};
static char _mssql_query_message[MSSQLDB_MSGSIZE]={0,};


/* _mssql.Connection class methods *******************************************/

static char _mssql_fetch_array_doc[] =
"fetch_array() -- get the whole result set.\n\n\
This method returns the whole result set from your queries:\n\
column names and types, then number of rows and rows themselves.\n\
";

// fetch_array
static PyObject *_mssql_fetch_array(_mssql_ConnectionObj *self, PyObject *args)
{
  PyObject *resultSet=NULL, *o=NULL;
  int retc, rows, i;
  // resultSet[ result(  header(...), recored(...) ),
  //         result(...),
  //         ...]
  if (!(resultSet = PyList_New(0))) {
    PyErr_SetString(_mssql_error,"Could not create fetch tuple");
    return NULL;
  }

  if(!self->connected) {
    PyErr_SetString(_mssql_error,"Not connected to any MS SQL server");
    return NULL;
  }

  // command executed correctly, get results information
  while((retc = dbresults(self->dbproc)) != NO_MORE_RESULTS) {
    PyObject *result=NULL, *header=NULL, *record=NULL;

    if (retc == FAIL)           // if error get out of loop
      break;

    if (!(result = PyTuple_New(3))) {
      PyErr_SetString(_mssql_error,"Could not create result tuple");
      return NULL;
    }

    header = GetHeaders(self->dbproc);       // print header data
    PyTuple_SET_ITEM(result,0,header);

    if (!(record = PyList_New(0))) {
      PyErr_SetString(_mssql_error,"Could not create record tuple");
      return NULL;
    }

    // loop on each row, until all read
    while((retc = dbnextrow(self->dbproc))!=NO_MORE_ROWS) {
      // if fail, then clear connection completely, just in case.
      if(retc == FAIL) {
        dbcancel(self->dbproc);
        //PyTuple_SET_ITEM(record,i++,Py_BuildValue("s",Py_BuildValue("s","\0")));
        o = PyString_FromString("");
        PyList_Append(record, o);
        Py_DECREF(o);
        break;
      } else {
        o = GetRow(self->dbproc);
        PyList_Append(record,o);
        Py_DECREF(o);
      }
    }

    rows=DBCOUNT(self->dbproc);
    i=0;
    // affected_rows
    PyTuple_SET_ITEM(result,1,PyInt_FromLong((long) rows));
    PyTuple_SET_ITEM(result,2,record);

    PyList_Append(resultSet,result);
    Py_DECREF(result);
  }   // end while(dbresults())

  return resultSet;
}

static char _mssql_query_doc[] =
"query(query_string) -- Send a query to the MS SQL Server.\n\n\
This method queries MS SQL Server and returns 1 if the query\n\
succeeded, 0 otherwise. You can then call stdmsg() to get\n\
reason for failing.\n\
";

// query
static PyObject *_mssql_query(_mssql_ConnectionObj *self, PyObject *args)
{
  char *Query = PyString_AsString(args);

  if(!self->connected) {
    PyErr_SetString(_mssql_error,"Not connected to any MS SQL server");
    return NULL;
  }

  *_mssql_query_error = *_mssql_query_message = '\0';

  dbcmd(self->dbproc,Query);

  if(dbsqlexec(self->dbproc)) {
    //dbfreebuf(self->dbproc);
    return PyInt_FromLong(1L);
  } else {
    //dbfreebuf(self->dbproc);
    return PyInt_FromLong(0L);
  }
}

static char _mssql_select_db_doc[] =
"select_db(dbname) -- Select a database and make it the current database.\n\n\
This function selects given database as the current one.\n\
";

// select_db
static PyObject *_mssql_select_db(_mssql_ConnectionObj *self, PyObject *args)
{
  char *Dbname = PyString_AsString(args);

  if(!self->connected) {
    PyErr_SetString(_mssql_error,"Not connected to any MS SQL server");
    return NULL;
  }

  *_mssql_query_error = *_mssql_query_message = '\0';

  if(dbuse(self->dbproc,Dbname))
    return PyInt_FromLong(1L);
  else
    return PyInt_FromLong(0L);
}

static char _mssql_close_doc[] =
"close() -- close connection to an MS SQL Server.\n\n\
This function tries to close the connection and free all memory\n\
it used. It returns ``None'' on success, NULL on failure.\n\
";

// close
static PyObject *_mssql_close(_mssql_ConnectionObj *self, PyObject *args)
{
  if(self->connected) {
    *_mssql_query_error = *_mssql_query_message = '\0';

    dbfreebuf(self->dbproc);
#ifdef MS_WINDOWS
    if(dbclose(self->dbproc) == FAIL) {
      PyErr_SetString(_mssql_error,"Failed to close and free the DB connection");
      return NULL;
    }
#else
    dbclose(self->dbproc);
#endif
    self->connected=0;
  } else {
    PyErr_SetString(_mssql_error,"Not connected to any MS SQL server");
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}


static char _mssql_errmsg_doc[] =
"errmsg() -- display DB-Lib or operating system error message.\n\n\
This method returns the error message from DB-Library or from\n\
operating system. ``None'' is returned if there was no error.\n\
It's both module method (this way you can find the reason for\n\
not being able to log in) and object method (erorrs occuring\n\
after connecting to an MS SQL Server.\n\n\
It's both _mssql module method and _mssql.Connection object method.\n\
";

static PyObject *_mssql_errmsg(_mssql_ConnectionObj *self, PyObject *args)
{
  if (*_mssql_query_error)
    return PyString_FromString(_mssql_query_error);
  else {
    Py_INCREF(Py_None);
    return Py_None;
  }
}

static char _mssql_stdmsg_doc[] =
"stdmsg() -- display SQL Server error message.\n\n\
This method returns the error message from SQL Server.\n\
You can safely ignore many of them, especially with severity<=10.\n\
If your query fails, this method can be a way to get info about\n\
what went wrong.\n\n\
It's both _mssql module method and _mssql.Connection object method.\n\
";

static PyObject *_mssql_stdmsg(_mssql_ConnectionObj *self, PyObject *args)
{
  if (*_mssql_query_message)
    return PyString_FromString(_mssql_query_message);
  else {
    Py_INCREF(Py_None);
    return Py_None;
  }
}


/* _mssql module methods *****************************************************/

//#define ERRDEBUG 1

int err_handler(DBPROCESS *dbproc, int severity, int dberr, int oserr,
		char *dberrstr, char *oserrstr)
{
#ifdef ERRDEBUG
  fprintf(stderr, "\n*** err_handler(dbproc=%lx, severity=%d, dberr=%d, \
oserr=%d, dberrstr='%s', oserrstr='%s'); DBDEAD(dbproc)=%d\n\n", dbproc,
severity, dberr, oserr, dberrstr, oserrstr, DBDEAD(dbproc));
#endif

  *_mssql_query_error = '\0';
  snprintf(_mssql_query_error, MSSQLDB_MSGSIZE, "DB-Lib error: %s\n",
		dberrstr);

  if ((oserr != DBNOERR) && (oserr != 0)) {
    /* get a textual representation of the error code */

#ifdef MS_WINDOWS
    HMODULE hModule = NULL; // default to system source
    LPSTR msg;
    DWORD buflen;
    DWORD fmtflags = FORMAT_MESSAGE_ALLOCATE_BUFFER |
		     FORMAT_MESSAGE_IGNORE_INSERTS |
		     FORMAT_MESSAGE_FROM_SYSTEM;

    if (oserr > NERR_BASE && oserr <= MAX_NERR) {
      hModule = LoadLibraryEx(TEXT("netmsg.dll"), NULL,
		   LOAD_LIBRARY_AS_DATAFILE);

      if (hModule != NULL)
        fmtflags |= FORMAT_MESSAGE_FROM_HMODULE;
    }

    if (buflen = FormatMessageA(fmtflags,
        hModule,	// module to get message from (NULL == system)
        oserr, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // default language
        (LPSTR) &msg, 0, NULL)) {
#else
#define EXCOMM 9
    char *msg = strerror(oserr);
#endif
      snprintf(_mssql_query_error + strlen(_mssql_query_error),
 		MSSQLDB_MSGSIZE - strlen(_mssql_query_error),
		"%s error during %s\n",
		(severity == EXCOMM) ? "Net-Lib" : "Operating system",
		oserrstr);
      snprintf(_mssql_query_error + strlen(_mssql_query_error),
		MSSQLDB_MSGSIZE - strlen(_mssql_query_error),
		"Error %d - %s", oserr, msg);

#ifdef MS_WINDOWS
      LocalFree(msg);

    // unload netmsg.dll
    if(hModule != NULL)
        FreeLibrary(hModule);
    }
#endif
  }

  return INT_CANCEL;  /* sigh FreeTDS sets DBDEAD on incorrect login! */
}

int msg_handler(DBPROCESS *dbproc, DBINT msgno, int msgstate, int severity,
		char *msgtext, char *srvname, char *procname, int line)
{
#ifdef ERRDEBUG
  fprintf(stderr, "\n+++ msg_handler(dbproc=%lx, msgno=%d, msgstate=%d, \
severity=%d, msgtext='%s', srvname='%s', procname='%s', line=%d\n\n", dbproc,
msgno, msgstate, severity, msgtext, srvname, procname, line);
#endif

  snprintf(_mssql_query_message, MSSQLDB_MSGSIZE,
	   "SQL Server message %ld, state %d, severity %d:\n%s\n",
	   (long)msgno, msgstate, severity, msgtext);

  return 0;
}


static char _mssql_connect_doc[] =
"connect(server,user,password) -- connect to an MS SQL Server.\n\
This method returns an instance of class _mssql.Connection.\n\n\
server   - is an instance of MS SQL server; it can be host name or\n\
           server identifier from freetds.conf. If you need to connect\n\
           to MS SQL listening on some strange port, just describe\n\
           it in freetds.conf.\n\
           Under Windows -- if you need to connect to a named\n\
           instance, just use the syntax 'hostname\\instancename'\n\
user     - user name to login as\n\
password - password to authenticate with\n\
";

static PyObject *_mssql_connect(_mssql_ConnectionObj *self, PyObject *args)
{
  _mssql_ConnectionObj *dbprochandle;
  LOGINREC  *login;         // allocate a DB-LIB login structure
  char *ServerName, *User, *Passwd;

#ifdef MS_WINDOWS
  char szAttribute[1024];
  char *p = szAttribute;
#endif

  if(!PyArg_ParseTuple(args,"sss:connect",&ServerName,&User,&Passwd)) {
    PyErr_SetString(_mssql_error,"Usage: connect(ServerName,User,Passwd)");
    return NULL;
  }

  *_mssql_query_error = *_mssql_query_message = '\0';

#ifdef MS_WINDOWS
  wsprintf(szAttribute, "DSN=pymssql;SERVER=%s;Trusted_connection=No;Network=DBMSSOCN;;", ServerName);
  while(*p) {
    if(*p == ';') *p = '\0';
    p++;
  }

  if(SQLConfigDataSource(NULL, ODBC_ADD_DSN, "SQL Server", szAttribute) == FALSE) {
    PyErr_SetString(_mssql_error,"DSN ADD ERROR");
    return NULL;
  }

  if(dbinit() == (char *)NULL) {
#else
  if(dbinit() == FAIL) {
#endif
    PyErr_SetString(_mssql_error,"Communication layer not loaded");
    return NULL;
  }

  dbsettime(30);	// set timeouts to 30 seconds
  login = dblogin();	// get login record from DB-LIB

#ifdef MS_WINDOWS
  dberrhandle((DBERRHANDLE_PROC)err_handler);
  dbmsghandle((DBMSGHANDLE_PROC)msg_handler);
#else
  dberrhandle(err_handler);
  dbmsghandle(msg_handler);
#endif

  DBSETLUSER(login, User);
  DBSETLPWD(login, Passwd);
  DBSETLAPP(login, "pymssql");
#ifdef MS_WINDOWS
  DBSETLVERSION(login, DBVER60);
#else
  DBSETLHOST(login, ServerName);
#endif

  dbprochandle = PyObject_NEW(_mssql_ConnectionObj, &_mssql_ConnectionObj_Type);
  if(dbprochandle == NULL) {
    PyErr_SetString(_mssql_error,"Could not create _mssql.Connection object");
    return NULL;
  }

  if((dbprochandle->dbproc=dbopen(login,ServerName)) == NULL) {
    PyErr_SetString(_mssql_error,"Could not connect to MS SQL Server");
    return NULL;
  }

  //dbsetopt(dbprochandle->dbproc, DBBUFFER, "2", -1);
  //dbsetopt(dbprochandle->dbproc, DBTEXTLIMIT, "8096", -1);

#ifdef MS_WINDOWS
  dbfreelogin(login);	// Frees a login record.
#else
  dbloginfree(login);	// Frees a login record.
#endif
  dbprochandle->connected=1;
  return (PyObject *)dbprochandle;
}


static void _mssql_ConnectionObj_dealloc(_mssql_ConnectionObj *self)
{
  if(self->connected) {
    PyObject *o =  _mssql_close(self, NULL);
    Py_XDECREF(o);
  }

  PyMem_Free((char *) self);
}


static PyObject * _mssql_ConnectionObj_repr(_mssql_ConnectionObj *self)
{
  char buf[300];
  if(self->connected)
    sprintf(buf,"<Open mssql connection at %lx>",(long)self);
  else
    sprintf(buf,"<Closed mssql connection at %lx>",(long)self);
  return PyString_FromString(buf);
}


static PyMethodDef _mssql_ConnectionObj_methods[] = {
  {"select_db",  (PyCFunction) _mssql_select_db,   METH_O,      _mssql_select_db_doc},
  {"query",      (PyCFunction) _mssql_query,       METH_O,      _mssql_query_doc},
  {"fetch_array",(PyCFunction) _mssql_fetch_array, METH_NOARGS, _mssql_fetch_array_doc},
  {"close",      (PyCFunction) _mssql_close,       METH_NOARGS, _mssql_close_doc},
  {"errmsg",     (PyCFunction) _mssql_errmsg,      METH_NOARGS, _mssql_errmsg_doc},
  {"stdmsg",     (PyCFunction) _mssql_stdmsg,      METH_NOARGS, _mssql_stdmsg_doc},
  {NULL,         NULL}
};


static PyMemberDef _mssql_ConnectionObj_members[] = {
  {"connected", T_INT, offsetof(_mssql_ConnectionObj, connected), READONLY,
   "True if the connection is open"},
  {NULL}
};


static char _mssql_ConnectionObj_Type_doc[] = {
"This object represents an MS SQL database connection. You can\n\
make queries and obtain results through a database connection.\n\
"};

static PyTypeObject _mssql_ConnectionObj_Type = {
#ifndef MS_WINDOWS
  PyObject_HEAD_INIT(&PyType_Type)
#else
  PyObject_HEAD_INIT(NULL)
#endif
  0,				/* ob_size           */
  "_mssql.Connection",		/* tp_name           */
  sizeof(_mssql_ConnectionObj),	/* tp_basicsize      */
  0,				/* tp_itemsize       */
  (destructor)_mssql_ConnectionObj_dealloc,	/* tp_dealloc     */
  0,				/* tp_print          */
  0,				/* tp_getattr        */
  0,				/* tp_setattr        */
  0,				/* tp_compare        */
  (reprfunc)_mssql_ConnectionObj_repr,		/* tp_repr        */
  0,				/* tp_as_number      */
  0,				/* tp_as_sequence    */
  0,				/* tp_as_mapping     */
  0,				/* tp_hash           */
  0,				/* tp_call           */
  0,				/* tp_str            */
  0,				/* tp_getattro       */
  0,				/* tp_setattro       */
  0,				/* tp_as_buffer      */
  Py_TPFLAGS_DEFAULT,		/* tp_flags          */
  _mssql_ConnectionObj_Type_doc,/* tp_doc            */
  0,				/* tp_traverse       */
  0,				/* tp_clear          */
  0,				/* tp_richcompare    */
  0,				/* tp_weaklistoffset */
  0,				/* tp_iter           */
  0,				/* tp_iternext       */
  _mssql_ConnectionObj_methods,	/* tp_methods        */
  _mssql_ConnectionObj_members,	/* tp_members        */
  0,				/* tp_getset         */
  0,				/* tp_base           */
  0,				/* tp_dict           */
  0,				/* tp_descr_get      */
  0,				/* tp_descr_set      */
  0,				/* tp_dictoffset     */
  0,				/* tp_init           */
  NULL,				/* tp_alloc          */
  NULL,				/* tp_new            */
  NULL,				/* tp_free Low-level free-memory routine */
  0,				/* tp_bases          */
  0,				/* tp_mro method resolution order */
  0,				/* tp_defined        */
};


static PyMethodDef _mssql_methods[] = {
  {"connect", (PyCFunction) _mssql_connect, METH_VARARGS, _mssql_connect_doc},
  {"errmsg",  (PyCFunction) _mssql_errmsg,  METH_NOARGS,  _mssql_errmsg_doc},
  {"stdmsg",  (PyCFunction) _mssql_stdmsg,  METH_NOARGS,  _mssql_stdmsg_doc},
  {NULL,      NULL}
};


#ifdef MS_WINDOWS
__declspec(dllexport)
#endif
void init_mssql(void)
{
  PyObject *m = Py_InitModule("_mssql", _mssql_methods);
  PyObject *d = PyModule_GetDict(m);

  /* if we initialize this at declaration, MSVC 7 issues the following warn:
   warning C4232: nonstandard extension used : 'tp_getattro': address of
   dllimport 'PyObject_GenericGetAttr' is not static, identity not guaranteed */
  _mssql_ConnectionObj_Type.tp_getattro = PyObject_GenericGetAttr;

  _mssql_error = PyErr_NewException("_mssql.error", NULL, NULL);
  PyDict_SetItemString(d, "error", _mssql_error);
  PyDict_SetItemString(d, "STRING", PyInt_FromLong(TYPE_STRING));
  PyDict_SetItemString(d, "BINARY", PyInt_FromLong(TYPE_BINARY));
  PyDict_SetItemString(d, "NUMBER", PyInt_FromLong(TYPE_NUMBER));
  PyDict_SetItemString(d, "DATETIME", PyInt_FromLong(TYPE_DATETIME));
  PyDict_SetItemString(d, "ROWID", PyInt_FromLong(TYPE_ROWID));
}

/* internal functions ********************************************************/

// dbproc -- The SQL Server connection structure
// col -- column size to get, 0 for all
int DetermineRowSize(DBPROCESS* dbproc, int col)
{
  int x,cols=0;        // counters
  int length=0;        // total length of column(row).
  int namelength;      // length of name of column
  int prlength;        // printable length
  char *name;          // pointer to column name

  if (!col)            // get number of columns
    cols = dbnumcols(dbproc);

  // count from 1 to numcols if col is 0, else x will = col only
  for(x=((col) ? col : 1);x<=((col) ? col : cols);x++) {
    // get column type, determine SQLCHAR converted length
    switch(dbcoltype(dbproc,x)) {
      case SQLNUMERIC:
      case SQLDECIMAL:
        {
#ifdef MS_WINDOWS
        DBCOL Col;
        Col.SizeOfStruct = sizeof(DBCOL);
        dbcolinfo(dbproc, CI_REGULAR, x, 0, &Col);
        prlength = Col.Precision + 2;
#else
        prlength = dbcollen(dbproc,x)+2;
#endif
        }
        break;

      case SQLBIT:		// The PR... values are found
        //prlength = PRBIT;	// in the SQLDB.H header file.
        prlength = 3;
        break;
      case SQLINT1:
        //prlength = PRINT1;
        prlength = 3;
        break;
      case SQLINT2:
        //prlength = PRINT2;
        prlength = 6;
        break;
      case SQLINT4:
        //prlength = PRINT4;
        prlength = 11;
        break;
      case SQLFLT8:
        //prlength = PRFLT8;
        prlength = 20;
        break;
      case SQLDATETIME:
        //prlength = PRDATETIME;
        prlength = 27;
        break;
      case SQLMONEY:
        //prlength = PRMONEY;
        prlength = 26;
        break;
      case SQLVARBINARY:	// VARBINARY IMAGE, and BINARY
      case SQLBINARY:		// convert to 2 times length
      case SQLIMAGE:
        prlength = dbcollen(dbproc,x)*2;
        break;
      default:
        // other types are maximum of actual column length
        prlength = dbcollen(dbproc,x);
        break;
    }

    // names may be longer than column so use name len if longer
    // of two. add one for space between columns
    name = (char*) dbcolname(dbproc, x);
    namelength = (name) ? strlen(name) : 0;

    if (prlength < namelength)
      length += namelength + 1;
    else
      length += prlength + 1;
  }

  return length;    // return the length of the field
}

PyObject *GetHeaders(DBPROCESS *dbproc)
{
  int x, cols, coltype, apicoltype;
  //int size;				// counters
  PyObject *headerSet=NULL;

  //size = DetermineRowSize(dbproc,0);		// get size of buffers
  //ptr = colnames = malloc(size+1);		// get name buffer
  //memset (colnames,' ',size);

  cols = dbnumcols(dbproc);			// get number of columns
  if (!(headerSet = PyTuple_New(cols))) {
    PyErr_SetString(_mssql_error,"Could not create column tuple");
    return NULL;
  }

  for(x = 1; x <= cols; x++) {			// loop on all columns
    PyObject *columnSet = NULL;
    if (!(columnSet = PyTuple_New(2))) {
      PyErr_SetString(_mssql_error,"Could not create tuple for column header details");
      return NULL;
    }

    //size = DetermineRowSize(dbproc,x);	// get size of this column
    //colname = (char *)dbcolname(dbproc,x);	// get column name
    //strncpy(ptr,colname,strlen(colname));	// copy name
    PyTuple_SET_ITEM(columnSet,0,Py_BuildValue("s",(char *)dbcolname(dbproc,x)));

    coltype = dbcoltype(dbproc,x);

    switch (coltype) {
      case SQLNUMERIC:
      case SQLDECIMAL:
      case SQLINT1:
      case SQLINT2:
      case SQLINT4:
      case SQLINTN:
      case SQLMONEY:
      case SQLMONEY4:
      case SQLMONEYN:
        apicoltype = TYPE_NUMBER;
        break;

      case SQLDATETIME:
      case SQLDATETIM4:
      case SQLDATETIMN:
        apicoltype = TYPE_DATETIME;
        break;

      case SQLVARBINARY:
      case SQLBINARY:
      case SQLIMAGE:
        apicoltype = TYPE_BINARY;
        break;

      case SQLVARCHAR:
      case SQLCHAR:
      case SQLTEXT:
        apicoltype = TYPE_STRING;
        break;

      default:
        apicoltype = TYPE_BINARY;
    }

    PyTuple_SET_ITEM(columnSet,1,Py_BuildValue("i",apicoltype));
    PyTuple_SET_ITEM(headerSet,x-1,columnSet);

    //ptr+=size;		// move to next position
  }

  //*ptr = '\0';		// null term both strings
  //printf("%s\n",colnames);	// print both strings
  //free(colnames);		// free both buffers

  return headerSet;
}

PyObject *GetRow(DBPROCESS *dbproc)
{
  int x,cols,coltype;		// counters
  LPBYTE data;			// column data pointer
  int size,datasize;
  char *datavals;
  long *datavali,intdata;
  double *datavalf,ddata;
  PyObject *record;

  datavali=&intdata;
  datavalf=&ddata;

  cols = dbnumcols(dbproc);	// get number of columns
  if (!(record = PyTuple_New(cols))) {
    PyErr_SetString(_mssql_error,"Not Created Record Tuple");
    return NULL;
  }

  for(x=1;x<=cols;x++) {	// do all columns
    coltype = dbcoltype(dbproc,x);
    size = DetermineRowSize(dbproc,x);
    data = (BYTE *)dbdata(dbproc,x);    // get pointer to column's data
    if(data == NULL) {		// if NULL, use "NULL"
      PyTuple_SET_ITEM(record,x-1,Py_BuildValue(""));
    } else {			// else have data, so convert to char
      datasize=dbdatlen(dbproc,x);
      if (datasize < size && (coltype == SQLNUMERIC ||
            coltype == SQLDECIMAL ||
            coltype == SQLINT1 ||
            coltype == SQLINT2 ||
            coltype == SQLINT4 ||
            coltype == SQLFLT8 ||
            coltype == SQLFLT4))
      {
        if(coltype==SQLFLT4 || coltype==SQLFLT8) {
          dbconvert(dbproc,coltype,data,datasize, SQLFLT8,(LPBYTE)datavalf,(DBINT)size-1);
          PyTuple_SET_ITEM(record,x-1,Py_BuildValue("d",*datavalf));
        } else {
          dbconvert(dbproc,coltype,data,datasize, SQLINT4,(LPBYTE)datavali,(DBINT)size-1);
          PyTuple_SET_ITEM(record,x-1,Py_BuildValue("i",*datavali));
        }
      } else {
        if(coltype==SQLDATETIME) {
          DBDATEREC di;
          char strdi[25]={0,};
          dbdatecrack(dbproc, &di, (DBDATETIME*)data);

          sprintf(strdi, "%02d-%02d-%02d %02d:%02d:%02d",
              di.year,
              di.month,		// see README.freetds for details
              di.day,
              di.hour,
              di.minute,
              di.second);
              //di.millisecond);

          PyTuple_SET_ITEM(record,x-1,Py_BuildValue("s",strdi));
        } else {
          datavals=malloc(datasize+1);
#ifdef MS_WINDOWS
          ZeroMemory(datavals,datasize+1);
#else
          memset(datavals,0,datasize+1);
#endif
          memmove(datavals,data,datasize);
          //printf("str %d\t- %s\n",datasize,datavals);
          PyTuple_SET_ITEM(record,x-1,Py_BuildValue("s",datavals));
          free(datavals);
        }
      }
    }
  }

  return record;             // done
}
